import re
import sys
import time
import json
from random import uniform
from .errors import RateLimitExceededError

import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

"""
contentful_management.utils
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements utilities.

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


logging.getLogger(__name__).addHandler(NullHandler())
log = logging.getLogger(__name__)


def unicode_class():
    """
    Returns the class that allows for unicode encoded strings
    depends on the Python version.
    """

    if sys.version_info[0] >= 3:
        return str
    return unicode  # noqa: F821


def string_class():
    """Returns the parent class for strings
    depends on the Python version."""
    if sys.version_info[0] >= 3:
        return str
    return basestring


def json_error_class():
    """Returns the class for JSON decode errors
    depends on the Python version."""
    if sys.version_info[0] >= 3 and sys.version_info[1] >= 5:
        return json.JSONDecodeError
    return ValueError


def snake_case(a_string):
    """
    Returns a snake-cased version of a string.

    :param a_string: any :class:`str` object.

    Usage:
        >>> snake_case('FooBar')
        "foo_bar"
    """

    partial = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', a_string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', partial).lower()


def camel_case(snake_str):
    """
    Returns a camel-cased version of a string.

    :param a_string: any :class:`str` object.

    Usage:
        >>> camel_case('foo_bar')
        "fooBar"
    """

    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])


def is_link(value):
    """
    Checks if value is link or not.

    :param value: any object.
    :return: Boolean
    :rtype: bool

    Usage:
        >>> is_link('foo')
        False
        >>> is_link({'sys': {'type': 'Link', 'id': 'foobar'}})
        True
    """

    return (
        isinstance(value, dict) and
        value.get('sys', {}).get('type', '') == 'Link'
    )


def is_link_array(value):
    """
    Checks if value is an array of links.

    :param value: any object.
    :return: Boolean
    :rtype: bool

    Usage:
        >>> is_link_array('foo')
        False
        >>> is_link_array([1, 2, 3])
        False
        >>> is_link([{'sys': {'type': 'Link', 'id': 'foobar'}}])
        True
    """

    if isinstance(value, list) and len(value) > 0:
        return is_link(value[0])
    return False


def base_path_for(resource_name):
    """
    Returns the path for a specified resource name.
    """

    return {
        'Role': 'roles',
        'Space': 'spaces',
        'Asset': 'assets',
        'Entry': 'entries',
        'Locale': 'locales',
        'Upload': 'uploads',
        'ApiKey': 'api_keys',
        'UIExtension': 'extensions',
        'ContentType': 'content_types',
        'Webhook': 'webhook_definitions',
        'PreviewApiKey': 'preview_api_keys',
        'SpaceMembership': 'space_memberships'
    }[resource_name]


def normalize_select(query):
    """
    If the query contains the :select operator, we enforce :sys properties.
    The SDK requires sys.type to function properly, but as other of our
    SDKs require more parts of the :sys properties, we decided that every
    SDK should include the complete :sys block to provide consistency
    accross our SDKs.
    """

    if 'select' not in query:
        return

    if isinstance(
            query['select'],
            str_type()):
        query['select'] = [s.strip() for s in query['select'].split(',')]

    query['select'] = [s for s
                       in query['select']
                       if not s.startswith('sys.')]

    if 'sys' not in query['select']:
        query['select'].append('sys')


def str_type():
    """
    Returns the correct string type for the current Python version.
    """

    if sys.version_info[0] >= 3:
        return str

    global basestring
    return basestring


class ConfigurationException(Exception):
    """
    Configuration error class.
    """

    pass


class retry_request(object):
    """
    Decorator to retry function calls in case they raise rate limit exceptions.
    """

    RATE_LIMIT_RESET_HEADER_KEY = 'x-contentful-ratelimit-reset'

    def __init__(self, client):
        self.client = client

    def __call__(self, http_call):
        def wrapper(url, query=None, **kwargs):
            exception = None
            for i in range(self.client.max_rate_limit_retries + 1):
                try:
                    return http_call(url, query, **kwargs)
                except RateLimitExceededError as error:
                    exception = error
                    reset_time = error.reset_time()

                    if reset_time > self.client.max_rate_limit_wait:
                        raise error

                    retry_message = 'Contentful API Rate Limit Hit! '
                    retry_message += "Retrying - Retries left: {0} ".format(
                        self.client.max_rate_limit_retries - i
                    )
                    retry_message += "- Time until reset (seconds): {0}".format(
                        reset_time
                    )
                    log.debug(retry_message)
                    time.sleep(reset_time * uniform(1.0, 1.2))
            raise exception
        return wrapper



from .client_proxy import ClientProxy
from .role import Role


"""
contentful_management.roles_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the RolesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class RolesProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles
    """

    @property
    def _resource_class(self):
        return Role

    def create(self, attributes=None, **kwargs):
        """
        Creates a role with given attributes.
        """

        return super(RolesProxy, self).create(None, attributes)



from .entry_resource_proxy import EntryResourceProxy
from .snapshots_proxy import SnapshotsProxy


"""
contentful_management.entry_snapshots_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EntrySnapshotsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EntrySnapshotsProxy(EntryResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots
    """
    def _resource_proxy_class(self):
        return SnapshotsProxy

    def create(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")



from .content_type_resource_proxy import ContentTypeResourceProxy
from .entries_proxy import EntriesProxy


"""
contentful_management.content_type_entries_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentTypeEntriesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentTypeEntriesProxy(ContentTypeResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries
    """

    def _resource_proxy_class(self):
        return EntriesProxy



from .client_proxy import ClientProxy
from .space_membership import SpaceMembership


"""
contentful_management.space_memberships_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpaceMembershipsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/space-memberships

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpaceMembershipsProxy(ClientProxy):
    @property
    def _resource_class(self):
        return SpaceMembership

    def create(self, attributes=None, **kwargs):
        """
        Creates a space membership with given attributes.
        """

        return super(SpaceMembershipsProxy, self).create(None, attributes)



from .client_proxy import ClientProxy
from .editor_interface import EditorInterface


"""
contentful_management.editor_interfaces_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EditorInterfacesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EditorInterfacesProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface
    """

    def __init__(self, client, space_id, environment_id=None, content_type_id=None):
        super(EditorInterfacesProxy, self).__init__(client, space_id, environment_id)
        self.content_type_id = content_type_id

    @property
    def _resource_class(self):
        return EditorInterface

    def all(self):
        """
        Gets the default editor interface.
        """

        return super(EditorInterfacesProxy, self).all()

    def find(self):
        """
        Gets the default editor interface.
        """

        return self.all()

    def default(self):
        """
        Gets the default editor interface.
        """

        return self.all()

    def create(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def _url(self, **kwargs):
        return self._resource_class.base_url(
            self.space_id,
            self.content_type_id,
            environment_id=self.environment_id
        )

    def __repr__(self):
        return "<{0} space_id='{1}' environment_id='{2}' content_type_id='{3}'>".format(
            self.__class__.__name__,
            self.space_id,
            self.environment_id,
            self.content_type_id
        )



from .content_type_resource_proxy import ContentTypeResourceProxy
from .editor_interfaces_proxy import EditorInterfacesProxy


"""
contentful_management.content_type_editor_interfaces_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentTypeEditorInterfacesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentTypeEditorInterfacesProxy(ContentTypeResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface
    """

    def _resource_proxy_class(self):
        return EditorInterfacesProxy

    def all(self):
        """
        Gets the default editor interface.
        """

        return self.proxy.all()

    def find(self, *args, **kwargs):
        """
        Gets the default editor interface.
        """

        return self.proxy.find()

    def default(self):
        """
        Gets the default editor interface.
        """

        return self.proxy.default()

    def create(self, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")



from .resource import Resource


"""
contentful_management.user
~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the User class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/users

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class User(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/users
    """

    def __init__(self, item, **kwargs):
        super(User, self).__init__(item, **kwargs)
        self.first_name = item.get('firstName', '')
        self.last_name = item.get('lastName', '')
        self.avatar_url = item.get('avatarUrl', '')
        self.email = item.get('email', '')
        self.activated = item.get('activated', False)
        self.sign_in_count = item.get('signInCount', 0)
        self.confirmed = item.get('confirmed', False)

    @classmethod
    def base_url(klass, **kwargs):
        """
        Returns the URI for the user.
        """

        return "users/me"

    def __repr__(self):
        return "<User[{0} {1}] email='{2}' activated={3} confirmed={4} sign_in_count={5}>".format(
            self.first_name,
            self.last_name,
            self.email,
            self.activated,
            self.confirmed,
            self.sign_in_count
        )



from .client_proxy import ClientProxy
from .ui_extension import UIExtension


"""
contentful_management.ui_extensions_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the UIExtensionsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/ui-extensions

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class UIExtensionsProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/ui-extensions
    """

    @property
    def _resource_class(self):
        return UIExtension



"""
contentful_management.entry_resource_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EntryResourceProxy class.

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EntryResourceProxy(object):
    """Base Class for Entry related Resource Proxies"""

    def __init__(self, client, space_id, environment_id, entry_id):
        self.proxy = self._resource_proxy_class()(client, space_id, environment_id, entry_id)

    def __repr__(self):
        return "<{0} space_id='{1}' environment_id='{2}' entry_id='{3}'>".format(
            self.__class__.__name__,
            self.proxy.space_id,
            self.proxy.environment_id,
            self.proxy.parent_resource_id
        )

    def _resource_proxy_class(self):
        raise Exception("Must implement")

    def all(self, query=None):
        """
        Gets all resources related to the current entry.
        """

        return self.proxy.all(query)

    def find(self, resource_id, query=None):
        """
        Finds a single resource by ID related to the current entry.
        """

        return self.proxy.find(resource_id, query)

    def create(self, resource_id=None, attributes=None):
        """
        Creates a resource with a given ID (optional) and attributes for the current entry.
        """

        return self.proxy.create(resource_id=resource_id, attributes=attributes)

    def delete(self, resource_id):
        """
        Deletes a resource by ID.
        """

        return self.proxy.delete(resource_id)



from .client_proxy import ClientProxy
from .upload import Upload
from .utils import str_type


"""
contentful_management.uploads_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the UploadsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/uploads

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class UploadsProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/uploads
    """

    @property
    def _resource_class(self):
        return Upload

    def all(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not supported")

    def create(self, file_or_path, **kwargs):
        """
        Creates an upload for the given file or path.
        """

        opened = False
        if isinstance(file_or_path, str_type()):
            file_or_path = open(file_or_path, 'rb')
            opened = True
        elif not getattr(file_or_path, 'read', False):
            raise Exception("A file or path to a file is required for this operation.")

        try:
            return self.client._post(
                self._url(),
                file_or_path,
                headers=self._resource_class.create_headers({}),
                file_upload=True
            )
        finally:
            if opened:
                file_or_path.close()

    def find(self, upload_id, **kwargs):
        """
        Finds an upload by ID.
        """

        return super(UploadsProxy, self).find(upload_id, file_upload=True)

    def delete(self, upload_id):
        """
        Deletes an upload by ID.
        """

        return super(UploadsProxy, self).delete(upload_id, file_upload=True)



from datetime import datetime
import dateutil.parser
from .resource import Resource


"""
contentful_management.webhook_call
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the WebhookCall class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-call-overview

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class WebhookCall(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-call-overview
    """

    def __init__(self, item, **kwargs):
        super(WebhookCall, self).__init__(item, **kwargs)
        self.request = item.get('request', {})
        self.response = item.get('response', {})
        self.status_code = item.get('statusCode', -1)
        self.errors = item.get('errors', [])
        self.event_type = item.get('eventType', '')
        self.url = item.get('url', '')
        try:
            self.request_at = dateutil.parser.parse(item.get('requestAt', ''))
        except:
            self.request_at = ''
        try:
            self.response_at = dateutil.parser.parse(item.get('responseAt', ''))
        except:
            self.response_at = ''

    @classmethod
    def base_url(klass, space_id, webhook_id, resource_id=None):
        """
        Returns the URI for the webhook call.
        """

        return "spaces/{0}/webhooks/{1}/calls/{2}".format(
            space_id,
            webhook_id,
            resource_id if resource_id is not None else ''
        )

    def __repr__(self):
        return "<WebhookCall[{0}] id='{1}' url='{2}' request_at='{3}' response_at='{4}'>".format(
            self.status_code,
            self.sys.get('id', ''),
            self.url,
            self.request_at.isoformat() if isinstance(self.request_at, datetime) else 'N/A',
            self.response_at.isoformat() if isinstance(self.request_at, datetime) else 'N/A'
        )



from .webhook_resource_proxy import WebhookResourceProxy
from .webhooks_health_proxy import WebhooksHealthProxy


"""
contentful_management.webhook_webhooks_health_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the WebhookWebhooksHealthProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-health

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class WebhookWebhooksHealthProxy(WebhookResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-health
    """

    def _resource_proxy_class(self):
        return WebhooksHealthProxy

    def find(self, *args, **kwargs):
        """
        Gets webhook health.
        """

        return self.all()



from .resource import FieldsResource, PublishResource, ArchiveResource, EnvironmentAwareResource
from .utils import is_link, is_link_array, snake_case
from .entry_snapshots_proxy import EntrySnapshotsProxy


"""
contentful_management.entry
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Entry class.

API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/reference/entries

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Entry(FieldsResource, PublishResource, ArchiveResource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/reference/entries
    """

    @classmethod
    def create_headers(klass, attributes):
        """
        Headers for entry creation.
        """

        if 'content_type_id' not in attributes:
            raise Exception("Content Type ID ('content_type_id') must be provided for this operation.")
        return {'x-contentful-content-type': attributes['content_type_id']}

    def __init__(self, *args, **kwargs):
        self.__CONTENT_TYPE__ = None
        super(Entry, self).__init__(*args, **kwargs)

    def snapshots(self):
        """
        Provides access to snapshot management methods for the given entry.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

        :return: :class:`EntrySnapshotsProxy <contentful_management.entry_snapshots_proxy.EntrySnapshotsProxy>` object.
        :rtype: contentful.entry_snapshots_proxy.EntrySnapshotsProxy

        Usage:

            >>> entry_snapshots_proxy = entry.snapshots()
            <EntrySnapshotsProxy space_id="cfexampleapi" environment_id="master" entry_id="nyancat">
        """
        return EntrySnapshotsProxy(self._client, self.sys['space'].id, self._environment_id, self.sys['id'])

    def update(self, attributes=None):
        """
        Updates the entry with attributes.
        """

        if attributes is None:
            attributes = {}

        attributes['content_type_id'] = self.sys['content_type'].id

        return super(Entry, self).update(attributes)

    def _coerce(self, value):
        if is_link(value):
            return self._build_link(value)
        elif is_link_array(value):
            return [self._build_link(link)
                    for link in value]
        return super(Entry, self)._coerce(value)

    def _missing_field_raw_id(self, name):
        for field in self._content_type().fields:
            if snake_case(field.id) == name:
                return field._real_id()

    def _is_missing_field(self, name):
        """
        Fields that are voided in the WebApp will be not returned in
        API responses, therefore we need to check if they are part of
        the content type to determine if they should or should not be
        serialized.
        """

        return self._missing_field_raw_id(name) is not None

    def _content_type(self):
        if self.__CONTENT_TYPE__ is None:
            self.__CONTENT_TYPE__ = self.sys['content_type'].resolve(self.sys['space'].id, environment_id=self._environment_id)
        return self.__CONTENT_TYPE__

    def _real_field_id_for(self, field_id):
        raw_field_id = super(Entry, self)._real_field_id_for(field_id)
        if raw_field_id is None:
            return self._missing_field_raw_id(field_id)
        return raw_field_id

    def __repr__(self):
        return "<Entry[{0}] id='{1}'>".format(
            self.sys['content_type'].sys.get('id', ''),
            self.sys.get('id', '')
        )



from .resource import Resource


"""
contentful_management.role
~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Role class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Role(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles
    """

    def __init__(self, item, **kwargs):
        super(Role, self).__init__(item, **kwargs)
        self.name = item.get('name', '')
        self.description = item.get('description', '')
        self.permissions = item.get('permissions', {})
        self.policies = item.get('policies', [])

    @classmethod
    def update_attributes_map(klass):
        """
        Defines keys and default values for non-generic attributes.
        """

        return {
            'name': '',
            'description': '',
            'permissions': {},
            'policies': []
        }

    def to_json(self):
        """
        Returns the JSON representation of the role.
        """

        result = super(Role, self).to_json()
        result.update({
            'name': self.name,
            'description': self.description,
            'permissions': self.permissions,
            'policies': self.policies
        })
        return result

    def __repr__(self):
        return "<Role[{0}] id='{1}'>".format(
            self.name,
            self.sys.get('id', '')
        )



from .client_proxy import ClientProxy
from .webhook_call import WebhookCall


"""
contentful_management.webhooks_call_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the WebhooksCallProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-call-overview

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class WebhooksCallProxy(ClientProxy):
    def __init__(self, client, space_id, webhook_id):
        super(WebhooksCallProxy, self).__init__(client, space_id)
        self.webhook_id = webhook_id

    @property
    def _resource_class(self):
        return WebhookCall

    def create(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def _url(self, resource_id='', **kwargs):
        return self._resource_class.base_url(
            self.space_id,
            self.webhook_id,
            resource_id=resource_id
        )

    def __repr__(self):
        return "<{0} space_id='{1}' webhook_id='{2}'>".format(
            self.__class__.__name__,
            self.space_id,
            self.webhook_id
        )



from .space_resource_proxy import SpaceResourceProxy
from .preview_api_keys_proxy import PreviewApiKeysProxy


"""
contentful_management.space_api_keys_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpacePreviewApiKeysProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys/preview-api-key/get-a-single-preview-api-key

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpacePreviewApiKeysProxy(SpaceResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys/preview-api-key/get-a-single-preview-api-key
    """

    def _resource_proxy_class(self):
        return PreviewApiKeysProxy

    def create(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")



from .webhook_resource_proxy import WebhookResourceProxy
from .webhooks_call_proxy import WebhooksCallProxy


"""
contentful_management.webhook_webhooks_call_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the WebhookWebhooksCallProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class WebhookWebhooksCallProxy(WebhookResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls
    """

    def _resource_proxy_class(self):
        return WebhooksCallProxy



"""
contentful_management.client_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ClientProxy class.

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ClientProxy(object):
    """
    Base class for proxies.
    """

    def __init__(self, client, space_id, environment_id=None):
        self.client = client
        self.space_id = space_id
        self.environment_id = environment_id

    def __repr__(self):
        return "<{0} space_id='{1}'{2}>".format(
            self.__class__.__name__,
            self.space_id,
            " environment_id='{0}'".format(self.environment_id) if self.environment_id is not None else ''
        )

    @property
    def _resource_class(self):
        raise Exception("Must Implement")

    def all(self, query=None):
        """
        Gets resource collection for _resource_class.
        """

        if query is None:
            query = {}
        return self.client._get(
            self._url(),
            query
        )

    def find(self, resource_id, query=None, **kwargs):
        """Gets a single resource."""

        if query is None:
            query = {}
        return self.client._get(
            self._url(resource_id),
            query,
            **kwargs
        )

    def create(self, resource_id=None, attributes=None):
        """
        Creates a resource with the given ID (optional) and attributes.
        """

        if attributes is None:
            attributes = {}

        result = None
        if not resource_id:
            result = self.client._post(
                self._url(resource_id),
                self._resource_class.create_attributes(attributes),
                headers=self._resource_class.create_headers(attributes)
            )
        else:
            result = self.client._put(
                self._url(resource_id),
                self._resource_class.create_attributes(attributes),
                headers=self._resource_class.create_headers(attributes)
            )

        return result

    def delete(self, resource_id, **kwargs):
        """
        Deletes a resource by ID.
        """

        return self.client._delete(self._url(resource_id), **kwargs)

    def _url(self, resource_id='', **kwargs):
        return self._resource_class.base_url(self.space_id, resource_id=resource_id, environment_id=self.environment_id, **kwargs)



from .resource import FieldsResource, PublishResource, ArchiveResource, EnvironmentAwareResource


"""
contentful_management.asset
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Asset class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Asset(FieldsResource, PublishResource, ArchiveResource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets
    """

    def url(self, **kwargs):
        """
        Returns a formatted URL for the asset's File
        with serialized parameters.

        Usage:
            >>> my_asset.url()
            "//images.contentful.com/spaces/foobar/..."
            >>> my_asset.url(w=120, h=160)
            "//images.contentful.com/spaces/foobar/...?w=120&h=160"
        """

        url = self.fields(self._locale()).get('file', {}).get('url', '')
        args = ['{0}={1}'.format(k, v) for k, v in kwargs.items()]

        if args:
            url += '?{0}'.format('&'.join(args))

        return url

    def process(self):
        """
        Calls the process endpoint for all locales of the asset.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets/asset-processing
        """

        for locale in self._fields.keys():
            self._client._put(
                "{0}/files/{1}/process".format(
                    self.__class__.base_url(
                        self.space.id,
                        self.id,
                        environment_id=self._environment_id
                    ),
                    locale
                ),
                {},
                headers=self._update_headers()
            )
        return self.reload()

    def __repr__(self):
        return "<Asset id='{0}' url='{1}'>".format(
            self.sys.get('id', ''),
            self.url()
        )



from .client_proxy import ClientProxy
from .preview_api_key import PreviewApiKey


"""
contentful_management.preview_api_keys_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the PreviewApiKeysProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class PreviewApiKeysProxy(ClientProxy):
    @property
    def _resource_class(self):
        return PreviewApiKey

    def create(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")



import dateutil.parser
from .resource import Resource


"""
contentful_management.PersonalAccessToken
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the PersonalAccessToken class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/personal-access-tokens

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class PersonalAccessToken(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/personal-access-tokens
    """

    def __init__(self, item, **kwargs):
        super(PersonalAccessToken, self).__init__(item, **kwargs)
        self.name = item.get('name', '')
        self.scopes = item.get('scopes', [])
        self.token = item.get('token', '')
        try:
            self.revoked_at = dateutil.parser.parse(item.get('revokedAt', ''))
        except:
            self.revoked_at = None

    @classmethod
    def base_url(klass, resource_id=None, **kwargs):
        """
        Returns the URI for the personal access token.
        """

        return "users/me/access_tokens/{0}".format(
            resource_id if resource_id is not None else ''
        )

    @property
    def is_revoked(self):
        return bool(self.revoked_at)

    def __repr__(self):
        return "<PersonalAccessToken[{0}] id='{1}' scopes=[{2}] revoked={3}>".format(
            self.name,
            self.sys.get('id', ''),
            ', '.join("'{0}'".format(s) for s in self.scopes),
            self.is_revoked
        )



from .utils import snake_case, camel_case


"""
contentful_management.content_type_field_validation
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentTypeFieldValidation class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types/content-type

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentTypeFieldValidation(object):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types/content-type
    """

    def __init__(self, validation_data):
        self.raw = validation_data
        self._data = {}
        for k, v in validation_data.items():
            self._data[snake_case(k)] = v

    def to_json(self):
        """
        Returns the JSON Representation of the content type field validation.
        """

        result = {}
        for k, v in self._data.items():
            result[camel_case(k)] = v
        return result

    def __getattr__(self, name):
        if name in self._data:
            return self._data[name]
        return super(ContentTypeFieldValidation, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name not in ['raw', '_data']:
            self._data[name] = value
        super(ContentTypeFieldValidation, self).__setattr__(name, value)

    def __repr__(self):
        return "<ContentTypeFieldValidation {0}>".format(
            " ".join(
                ["{0}='{1}'".format(k, v)
                 for k, v in self._data.items()]
            )
        )



try:
    import simplejson as json
except ImportError:
    import json

import dateutil.parser
from .utils import unicode_class

"""
contentful_management.content_type_field_types
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the field coercion classes.

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class BasicField(object):
    """
    Base coercion class.
    """

    def __init__(self, items=None):
        self._items = items

    def coerce(self, value):
        """
        Just returns the value.
        """

        return value

    def __repr__(self):
        return "<{0}>".format(
            self.__class__.__name__
        )


class SymbolField(BasicField):
    """
    Symbol coercion class.
    """

    def coerce(self, value):
        """
        Coerces value to str.
        """

        return unicode_class()(value)


class TextField(SymbolField):
    """
    Text coercion class.
    """
    pass


class IntegerField(BasicField):
    """
    Integer coercion class.
    """

    def coerce(self, value):
        """
        Coerces value to int.
        """

        return int(value) if value is not None else None


class NumberField(BasicField):
    """
    Number coercion class.
    """

    def coerce(self, value):
        """
        Coerces value to float.
        """

        return float(value) if value is not None else None


class DateField(BasicField):
    """
    Date coercion class.
    """

    def coerce(self, value):
        """
        Coerces ISO8601 date to :class:`datetime.datetime` object.
        """

        return dateutil.parser.parse(value)


class BooleanField(BasicField):
    """
    Boolean coercion class.
    """

    def coerce(self, value):
        """
        Coerces value to boolean.
        """

        return bool(value)


class LinkField(BasicField):
    """
    LinkField

    Nothing should be done here as include resolution is handled within
    entries due to depth handling (explained within Entry).
    Only present as a placeholder for proper resolution within ContentType.
    """
    pass


class ArrayField(BasicField):
    """
    Array coercion class.

    Coerces items in collection with it's proper coercion class.
    """

    def __init__(self, items=None):
        super(ArrayField, self).__init__(items)
        self._coercion = self._get_coercion()

    def coerce(self, value):
        """
        Coerces array items with proper coercion.
        """

        result = []
        for v in value:
            result.append(self._coercion.coerce(v))
        return result

    def _get_coercion(self):
        return globals()["{0}Field".format(self._items.get('type'))]()


class ObjectField(BasicField):
    """
    Object coercion class.
    """

    def coerce(self, value):
        """
        Coerces value to JSON.
        """

        return json.loads(json.dumps(value))


class LocationField(BasicField):
    """
    Location coercion class.
    """

    def coerce(self, value):
        """
        Coerces value to location hash.
        """

        return {
            'lat': float(value.get('lat', value.get('latitude'))),
            'lon': float(value.get('lon', value.get('longitude')))
        }



from .resource import Resource, EnvironmentAwareResource


"""
contentful_management.editor_interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EditorInterface class.

API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/reference/editor-interfaces

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EditorInterface(Resource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/reference/editor-interfaces
    """

    def __init__(self, item, **kwargs):
        super(EditorInterface, self).__init__(item, **kwargs)
        self.controls = item.get('controls', [])

    @classmethod
    def base_url(self, space_id, content_type_id, environment_id=None, **kwargs):
        """
        Returns the URI for the editor interface.
        """

        return "spaces/{0}{1}/content_types/{2}/editor_interface".format(
            space_id,
            '/environments/{0}'.format(environment_id) if environment_id is not None else '',
            content_type_id
        )

    @classmethod
    def update_attributes_map(klass):
        """
        Attributes for object mapping.
        """

        return {'controls': []}

    def to_json(self):
        """
        Returns the JSON representation of the editor interface.
        """

        result = super(EditorInterface, self).to_json()
        result.update({'controls': self.controls})
        return result

    def _update_url(self):
        return self.__class__.base_url(
            self.space.id,
            self.content_type.id,
            environment_id=self._environment_id
        )

    def __repr__(self):
        return "<EditorInterface id='{0}'>".format(
            self.sys.get('id', '')
        )



from .client_proxy import ClientProxy
from .space import Space


"""
contentful_management.spaces_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpacesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/spaces

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpacesProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/spaces
    """

    def __init__(self, client):
        super(SpacesProxy, self).__init__(client, None)

    def __repr__(self):
        return "<SpacesProxy>"

    @property
    def _resource_class(self):
        return Space

    def all(self, query=None, **kwargs):
        """
        Gets all spaces.
        """

        return super(SpacesProxy, self).all(query=query)

    def find(self, space_id, query=None, **kwargs):
        """
        Gets a space by ID.
        """

        try:
            self.space_id = space_id
            return super(SpacesProxy, self).find(space_id, query=query)
        finally:
            self.space_id = None

    def create(self, attributes=None, **kwargs):
        """
        Creates a space with given attributes.
        """

        if attributes is None:
            attributes = {}
        if 'default_locale' not in attributes:
            attributes['default_locale'] = self.client.default_locale

        return super(SpacesProxy, self).create(resource_id=None, attributes=attributes)

    def delete(self, space_id):
        """
        Deletes a space by ID.
        """

        try:
            self.space_id = space_id
            return super(SpacesProxy, self).delete(space_id)
        finally:
            self.space_id = None



from .resource import Resource, EnvironmentAwareResource


"""
contentful_management.snapshot
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Snapshot class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Snapshot(Resource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots
    """

    def __init__(self, item, **kwargs):
        super(Snapshot, self).__init__(item, **kwargs)

        from .entry import Entry
        from .content_type import ContentType
        entity_type = {
            'Entry': Entry,
            'ContentType': ContentType
        }

        if self.sys['snapshot_entity_type'] not in entity_type:
            raise Exception("Object '{0}' not buildable".format(self.sys['snapshot_entity_type']))

        self.snapshot = entity_type[self.sys['snapshot_entity_type']](item['snapshot'], **kwargs)

    @classmethod
    def base_url(klass, space_id, parent_resource_id, resource_url='entries', resource_id=None, environment_id=None):
        """
        Returns the URI for the snapshot.
        """

        return "spaces/{0}{1}/{2}/{3}/snapshots/{4}".format(
            space_id,
            '/environments/{0}'.format(environment_id) if environment_id is not None else '',
            resource_url,
            parent_resource_id,
            resource_id if resource_id is not None else ''
        )

    def to_json(self):
        """
        Returns the JSON representation of the snapshot.
        """

        result = super(Snapshot, self).to_json()
        result.update({
            'snapshot': self.snapshot.to_json(),
        })
        return result

    def __repr__(self):
        return "<Snapshot[{0}] id='{1}'>".format(
            self.sys.get('snapshot_entity_type', ''),
            self.sys.get('id', '')
        )

    def save(self):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def update(self, _attributes=None):
        """
        Not supported.
        """

        raise Exception("Not Supported")



from .resource import Resource, EnvironmentAwareResource


"""
contentful_management.locale
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Locale class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Locale(Resource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales
    """

    def __init__(self, item, **kwargs):
        super(Locale, self).__init__(item, **kwargs)
        self.code = item.get('code', '')
        self.name = item.get('name', '')
        self.fallback_code = item.get('fallbackCode', '')
        self.default = item.get('default', False)
        self.optional = item.get('optional', True)
        self.content_delivery_api = item.get('contentDeliveryApi', True)
        self.content_management_api = item.get('contentManagementApi', True)

    @classmethod
    def update_attributes_map(klass):
        """
        Attributes for object mapping.
        """

        return {
            'code': '',
            'name': '',
            'fallback_code': '',
            'default': False,
            'optional': True,
            'content_delivery_api': True,
            'content_management_api': True
        }

    def to_json(self):
        """
        Returns the JSON representation of the locale.
        """

        result = super(Locale, self).to_json()
        result.update({
            'code': self.code,
            'name': self.name,
            'fallbackCode': self.fallback_code,
            'optional': self.optional,
            'contentDeliveryApi': self.content_delivery_api,
            'contentManagementApi': self.content_management_api
        })
        return result

    def __repr__(self):
        return "<Locale[{0}] code='{1}' default={2}>".format(
            self.name,
            self.code,
            self.default
        )



from .resource import Resource, PublishResource, EnvironmentAwareResource
from .content_type_field import ContentTypeField
from .content_type_entries_proxy import ContentTypeEntriesProxy
from .content_type_snapshots_proxy import ContentTypeSnapshotsProxy
from .content_type_editor_interfaces_proxy import ContentTypeEditorInterfacesProxy


"""
contentful_management.content_type
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentType class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentType(Resource, PublishResource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types
    """

    def __init__(self, item, **kwargs):
        super(ContentType, self).__init__(item, **kwargs)
        self.name = item.get('name', '')
        self.description = item.get('description', '')
        self.display_field = item.get('displayField', '')
        self.fields = [ContentTypeField(field)
                       for field in item.get('fields', [])]

    @classmethod
    def base_url(klass, space_id, resource_id=None, public=False, environment_id=None, **kwargs):
        """
        Returns the URI for the content type.
        """

        if public:
            environment_slug = ""
            if environment_id is not None:
                environment_slug = "/environments/{0}".format(environment_id)
            return "spaces/{0}{1}/public/content_types".format(space_id, environment_slug)
        return super(ContentType, klass).base_url(
            space_id,
            resource_id=resource_id,
            environment_id=environment_id,
            **kwargs
        )

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """
        Attributes for content type creation.
        """

        result = super(ContentType, klass).create_attributes(attributes, previous_object)

        if 'fields' not in result:
            result['fields'] = []
        return result

    @classmethod
    def update_attributes_map(klass):
        """
        Attributes for object mapping.
        """

        return {
            'name': '',
            'description': '',
            'display_field': '',
            'fields': []
        }

    def to_json(self):
        """
        Returns the JSON representation of the content type.
        """

        result = super(ContentType, self).to_json()
        result.update({
            'name': self.name,
            'description': self.description,
            'displayField': self.display_field,
            'fields': [f.to_json() for f in self.fields]
        })
        return result

    def entries(self):
        """
        Provides access to entry management methods for the given content type.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries

        :return: :class:`ContentTypeEntriesProxy <contentful_management.content_type_entries_proxy.ContentTypeEntriesProxy>` object.
        :rtype: contentful.content_type_entries_proxy.ContentTypeEntriesProxy

        Usage:

            >>> content_type_entries_proxy = content_type.entries()
            <ContentTypeEntriesProxy space_id="cfexampleapi" environment_id="master" content_type_id="cat">
        """
        return ContentTypeEntriesProxy(self._client, self.space.id, self._environment_id, self.id)

    def editor_interfaces(self):
        """
        Provides access to editor interface management methods for the given content type.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface

        :return: :class:`ContentTypeEditorInterfacesProxy <contentful_management.content_type_editor_interfaces_proxy.ContentTypeEditorInterfacesProxy>` object.
        :rtype: contentful.content_type_editor_interfaces_proxy.ContentTypeEditorInterfacesProxy

        Usage:

            >>> content_type_editor_interfaces_proxy = content_type.editor_interfaces()
            <ContentTypeEditorInterfacesProxy space_id="cfexampleapi" environment_id="master" content_type_id="cat">
        """
        return ContentTypeEditorInterfacesProxy(self._client, self.space.id, self._environment_id, self.id)

    def snapshots(self):
        """
        Provides access to snapshot management methods for the given content type.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots/content-type-snapshots-collection

        :return: :class:`ContentTypeSnapshotsProxy <contentful_management.content_type_snapshots_proxy.ContentTypeSnapshotsProxy>` object.
        :rtype: contentful.content_type_snapshots_proxy.ContentTypeSnapshotsProxy

        Usage:

            >>> content_type_snapshots_proxy = content_type.entries()
            <ContentTypeSnapshotsProxy space_id="cfexampleapi" environment_id="master" content_type_id="cat">
        """
        return ContentTypeSnapshotsProxy(self._client, self.space.id, self._environment_id, self.id)

    def __repr__(self):
        return "<ContentType[{0}] id='{1}'>".format(
            self.name,
            self.sys.get('id', '')
        )



from .client_proxy import ClientProxy
from .entry import Entry
from .utils import normalize_select


"""
contentful_management.entries_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EntriesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EntriesProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries
    """

    def __init__(self, client, space_id, environment_id=None, content_type_id=None):
        super(EntriesProxy, self).__init__(client, space_id, environment_id=environment_id)
        self.content_type_id = content_type_id

    @property
    def _resource_class(self):
        return Entry

    def all(self, query=None):
        """
        Gets all entries of a space.
        """

        if query is None:
            query = {}

        if self.content_type_id is not None:
            query['content_type'] = self.content_type_id

        normalize_select(query)

        return super(EntriesProxy, self).all(query=query)

    def find(self, entry_id, query=None):
        """
        Gets a single entry by ID.
        """

        if query is None:
            query = {}

        if self.content_type_id is not None:
            query['content_type'] = self.content_type_id

        normalize_select(query)

        return super(EntriesProxy, self).find(entry_id, query=query)

    def create(self, resource_id=None, attributes=None, **kwargs):
        """
        Creates an entry with a given ID (optional) and attributes.
        """

        if self.content_type_id is not None:
            if attributes is None:
                attributes = {}
            attributes['content_type_id'] = self.content_type_id

        return super(EntriesProxy, self).create(resource_id=resource_id, attributes=attributes)



from .utils import snake_case
from .content_type_field_types import *  # noqa: F403, F401
from .content_type_field_validation import ContentTypeFieldValidation

"""
contentful_management.content_type_field
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentTypeField class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types/content-type

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentTypeField(object):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types/content-type
    """

    def __init__(self, field_data):
        self.raw = field_data
        self.id = snake_case(field_data.get('id', ''))
        self.name = field_data.get('name', '')
        self.type = field_data.get('type', '')
        self.link_type = field_data.get('linkType', '')
        self.items = field_data.get('items', {})
        self.localized = field_data.get('localized', False)
        self.omitted = field_data.get('omitted', False)
        self.required = field_data.get('required', False)
        self.disabled = field_data.get('disabled', False)
        self.validations = [ContentTypeFieldValidation(v)
                            for v in field_data.get('validations', [])]
        self._coercion = self._get_coercion()

    def to_json(self):
        """
        Returns the JSON Representation of the content type field.
        """

        result = {
            'name': self.name,
            'id': self._real_id(),
            'type': self.type,
            'localized': self.localized,
            'omitted': self.omitted,
            'required': self.required,
            'disabled': self.disabled,
            'validations': [v.to_json() for v in self.validations]
        }

        if self.type == 'Array':
            result['items'] = self.items

        if self.type == 'Link':
            result['linkType'] = self.link_type

        return result

    def coerce(self, value):
        """
        Coerces the value to the proper type.
        """

        return self._coercion.coerce(value)

    def _get_coercion(self):
        """
        Gets the proper coercion type.
        """

        return globals()["{0}Field".format(self.type)](self.items)

    def _real_id(self):
        """Gets the real field ID."""

        return self.raw['id']

    def __repr__(self):
        return "<ContentTypeField[{0}] id='{1}' type='{2}'>".format(
            self.name,
            self.id,
            self.type
        )



from .client_proxy import ClientProxy
from .asset import Asset
from .utils import normalize_select


"""
contentful_management.assets_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the AssetsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class AssetsProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets
    """

    @property
    def _resource_class(self):
        return Asset

    def all(self, query=None, **kwargs):
        """
        Gets all assets of a space.
        """

        if query is None:
            query = {}

        normalize_select(query)

        return super(AssetsProxy, self).all(query, **kwargs)

    def find(self, asset_id, query=None, **kwargs):
        """
        Gets a single asset by ID.
        """

        if query is None:
            query = {}

        normalize_select(query)

        return super(AssetsProxy, self).find(asset_id, query=query, **kwargs)



import dateutil.parser

from datetime import datetime

from .utils import snake_case
from .utils import camel_case
from .utils import base_path_for


"""
contentful_management.resource
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Resource, FieldResource, PublishResource, ArchiveResource and Link classes.

API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/introduction/common-resource-attributes

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Resource(object):
    """
    Base resource class.

    Implements common resource attributes.

    API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/introduction/common-resource-attributes
    """

    def __init__(self, item, default_locale='en-US', client=None):
        self.raw = item
        self.default_locale = default_locale
        self._client = client
        self.sys = self._hydrate_sys(item)

    @classmethod
    def base_url(klass, space_id='', resource_id=None, environment_id=None, **kwargs):
        """
        Returns the URI for the resource.
        """

        url = "spaces/{0}".format(
            space_id)

        if environment_id is not None:
            url = url = "{0}/environments/{1}".format(url, environment_id)

        url = "{0}/{1}".format(
            url,
            base_path_for(klass.__name__)
        )

        if resource_id:
            url = "{0}/{1}".format(url, resource_id)

        return url

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """
        Attributes for resource creation.
        """

        result = {}

        if previous_object is not None:
            result = {k: v for k, v in previous_object.to_json().items() if k != 'sys'}

        result.update(attributes)

        return result

    @classmethod
    def create_headers(klass, attributes):
        """
        Headers for resource creation.
        """

        return {}

    @classmethod
    def update_attributes_map(klass):
        """
        Defines keys and default values for non-generic attributes.
        """

        return {}

    def delete(self):
        """
        Deletes the resource.
        """

        return self._client._delete(
            self.__class__.base_url(
                self.sys['space'].id,
                self.sys['id'],
                environment_id=self._environment_id
            )
        )

    def update(self, attributes=None):
        """
        Updates the resource with attributes.
        """

        if attributes is None:
            attributes = {}

        headers = self.__class__.create_headers(attributes)
        headers.update(self._update_headers())

        result = self._client._put(
            self._update_url(),
            self.__class__.create_attributes(attributes, self),
            headers=headers
        )

        self._update_from_resource(result)

        return self

    def save(self):
        """
        Saves the current state of the resource.
        """

        return self.update()

    def reload(self, result=None):
        """
        Reloads the resource.
        """

        if result is None:
            result = self._client._get(
                self.__class__.base_url(
                    self.sys['space'].id,
                    self.sys['id'],
                    environment_id=self._environment_id
                )
            )

        self._update_from_resource(result)

        return self

    def to_link(self):
        """
        Returns a link for the resource.
        """

        link_type = self.link_type if self.type == 'Link' else self.type

        return Link({'sys': {'linkType': link_type, 'id': self.sys.get('id')}}, client=self._client)

    def to_json(self):
        """
        Returns the JSON representation of the resource.
        """

        result = {
            'sys': {}
        }
        for k, v in self.sys.items():
            if k in ['space', 'content_type', 'created_by',
                     'updated_by', 'published_by']:
                v = v.to_json()
            if k in ['created_at', 'updated_at', 'deleted_at',
                     'first_published_at', 'published_at', 'expires_at']:
                v = v.isoformat()
            result['sys'][camel_case(k)] = v

        return result

    def _hydrate_sys(self, item):
        sys = {}
        for k, v in item['sys'].items():
            if k in ['space', 'contentType', 'createdBy',
                     'updatedBy', 'publishedBy', 'environment']:
                v = self._build_link(v)
            if k in ['createdAt', 'updatedAt', 'deletedAt',
                     'firstPublishedAt', 'publishedAt', 'expiresAt']:
                v = dateutil.parser.parse(v)
            sys[snake_case(k)] = v
        return sys

    def _build_link(self, link):
        return Link(link, client=self._client)

    def _update_headers(self):
        return {'x-contentful-version': str(self.sys['version'])}

    def _update_url(self):
        return self.__class__.base_url(
                self.sys['space'].id,
                self.sys['id'],
                environment_id=self._environment_id
            )

    def _update_from_resource(self, other):
        if hasattr(other, 'sys'):
            self.sys = other.sys
        for attr, default in self.__class__.update_attributes_map().items():
            value = getattr(other, attr, default)
            self_value = getattr(self, attr)
            if value == default and value != self_value:
                value = self_value
            setattr(self, attr, value)

    @property
    def _environment_id(self):
        """
        Returns the Environment ID.
        """
        try:
            return super(Resource, self)._environment_id
        except AttributeError:
            # In Resources which do not inherit EnvironmentAwareResource an AttributeError will happen
            return None

    def __getattr__(self, name):
        if name in self.sys:
            return self.sys[name]
        raise AttributeError(
            "'{0}' object has no attribute '{1}'".format(
                self.__class__.__name__,
                name
            )
        )


class FieldsResource(Resource):
    """
    Fields resource class.

    Implements locale handling for resource fields.
    """

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """
        Attributes for resource creation.
        """

        if 'fields' not in attributes:
            if previous_object is None:
                attributes['fields'] = {}
            else:
                attributes['fields'] = previous_object.to_json()['fields']
        return {'fields': attributes['fields']}

    def __init__(self, item, **kwargs):
        super(FieldsResource, self).__init__(item, **kwargs)
        self._fields = self._hydrate_fields(item)

    def fields(self, locale=None):
        """
        Get fields for a specific locale.

        :param locale: (optional) Locale to fetch, defaults to default_locale.
        """

        if locale is None:
            locale = self._locale()
        return self._fields.get(locale, {})

    def fields_with_locales(self):
        """
        Get fields with locales per field.
        """

        result = {}
        for locale, fields in self._fields.items():
            for k, v in fields.items():
                real_field_id = self._real_field_id_for(k)
                if real_field_id not in result:
                    result[real_field_id] = {}
                result[real_field_id][locale] = self._serialize_value(v)
        return result

    def to_json(self):
        """
        Returns the JSON Representation of the resource.
        """

        result = super(FieldsResource, self).to_json()
        result['fields'] = self.fields_with_locales()
        return result

    @property
    def locale(self):
        """
        Returns the resource locale.
        """

        return self.sys.get('locale', None)

    def _real_field_id_for(self, field_id):
        for raw_field_id in self.raw['fields'].keys():
            if snake_case(raw_field_id) == field_id:
                return raw_field_id

    def _serialize_value(self, value):
        if isinstance(value, Resource):
            return value.to_link().to_json()
        elif isinstance(value, list) and value:
            if isinstance(value[0], Resource):
                return [resource.to_link().to_json() for resource in value]
        elif isinstance(value, datetime):
            return value.isoformat()
        return value

    def _hydrate_fields(self, item):
        fields = {}

        if 'fields' not in item:
            return fields

        for k, locales in item['fields'].items():
            for locale, v in locales.items():
                if locale not in fields:
                    fields[locale] = {}
                fields[locale][snake_case(k)] = self._coerce(v)

        return fields

    def _coerce(self, value):
        return value

    def _locale(self):
        return self.locale or self.__dict__['default_locale']

    def _update_from_resource(self, other):
        super(FieldsResource, self)._update_from_resource(other)
        if hasattr(other, '_fields'):
            self._fields = other._fields

    def __getattr__(self, name):
        locale = self._locale()
        if name in self._fields.get(locale, {}):
            return self._fields[locale][name]
        return super(FieldsResource, self).__getattr__(name)

    def __setattr__(self, name, value):
        if name not in ['raw', 'sys', 'default_locale',
                        '_client', '_fields', '__CONTENT_TYPE__']:
            locale = self._locale()
            if (name in self._fields.get(locale, {}) or
                    self._is_missing_field(name)):
                if locale not in self._fields:
                    self._fields[locale] = {}
                self._fields[locale][name] = value
                return self._fields[locale][name]
        return super(FieldsResource, self).__setattr__(name, value)

    def _is_missing_field(self, name):
        """
        By default, fields not appearing on responses are considered
        as object meta-data, and they will not be added to `_fields`,
        making them not part of the serialization when sent back to
        the API for saving.
        """

        return False


class PublishResource(object):
    """
    Allows for resource publish/unpublish.
    """

    @property
    def is_published(self):
        """
        Checks if resource is published.
        """

        return bool(self.sys.get('published_at', False))

    def publish(self):
        """
        Publishes the resource.
        """

        result = self._client._put(
            "{0}/published".format(
                self.__class__.base_url(
                    self.sys['space'].id,
                    self.sys['id'],
                    environment_id=self._environment_id
                ),
            ),
            {},
            headers=self._update_headers()
        )

        return self.reload(result)

    def unpublish(self):
        """
        Unpublishes the resource.
        """

        self._client._delete(
            "{0}/published".format(
                self.__class__.base_url(
                    self.sys['space'].id,
                    self.sys['id'],
                    environment_id=self._environment_id
                ),
            ),
            headers=self._update_headers()
        )

        return self.reload()


class ArchiveResource(object):
    """
    Allows for resource archive/unarchive.
    """

    @property
    def is_archived(self):
        """
        Checks if Resource is archived.
        """

        return bool(self.sys.get('archived_version', False))

    def archive(self):
        """
        Archives the resource.
        """

        self._client._put(
            "{0}/archived".format(
                self.__class__.base_url(
                    self.sys['space'].id,
                    self.sys['id'],
                    environment_id=self._environment_id
                ),
            ),
            {},
            headers=self._update_headers()
        )

        return self.reload()

    def unarchive(self):
        """
        Unarchives the resource.
        """

        self._client._delete(
            "{0}/archived".format(
                self.__class__.base_url(
                    self.sys['space'].id,
                    self.sys['id'],
                    environment_id=self._environment_id
                ),
            ),
            headers=self._update_headers()
        )

        return self.reload()


class EnvironmentAwareResource(object):
    """
    Allows environment aware resources to resolve the environment ID.
    """

    @property
    def _environment_id(self):
        """
        Returns the Environment ID.
        """

        environment = self.sys.get('environment', None)

        if environment is not None:
            return environment.id
        return 'master'


class Link(Resource):
    """
    Link Class

    API reference: https://www.contentful.com/developers/docs/references/content-delivery-api/#/reference/links
    """

    def resolve(self, space_id=None, environment_id=None):
        """
        Resolves link to a specific resource.
        """

        proxy_method = getattr(
            self._client,
            base_path_for(self.link_type)
        )
        if self.link_type == 'Space':
            return proxy_method().find(self.id)
        elif environment_id is not None:
            return proxy_method(space_id, environment_id).find(self.id)
        else:
            return proxy_method(space_id).find(self.id)

    def to_json(self):
        """
        Returns the JSON representation of the link.
        """

        return {
            'sys': {
                'type': 'Link',
                'linkType': self.sys.get('link_type'),
                'id': self.sys.get('id')
            }
        }

    def __repr__(self):
        return "<Link[{0}] id='{1}'>".format(
            self.link_type,
            self.id
        )



from .environment_resource_proxy import EnvironmentResourceProxy
from .locales_proxy import LocalesProxy


"""
contentful_management.environment_locales_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EnvironmentLocalesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EnvironmentLocalesProxy(EnvironmentResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries
    """

    def _resource_proxy_class(self):
        return LocalesProxy



from .resource import Resource


"""
contentful_management.space_membership
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpaceMembership class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/space-memberships

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpaceMembership(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/space-memberships
    """

    def __init__(self, item, **kwargs):
        super(SpaceMembership, self).__init__(item, **kwargs)
        self.admin = item.get('admin', False)
        self.roles = item.get('roles', [])
        self.user = item.get('user', None)

    def to_json(self):
        """
        Returns the JSON representation of the space membership.
        """

        result = super(SpaceMembership, self).to_json()
        result.update({
            'admin': self.admin,
            'roles': self.roles
        })
        return result

    def __repr__(self):
        return "<SpaceMembership id='{0}' admin={1}>".format(
            self.sys.get('id', ''),
            self.admin
        )



from .client import Client  # noqa: F401
from .entry import Entry  # noqa: F401
from .asset import Asset  # noqa: F401
from .space import Space  # noqa: F401
from .locale import Locale  # noqa: F401
from .api_key import ApiKey  # noqa: F401
from .preview_api_key import PreviewApiKey  # noqa: F401
from .editor_interface import EditorInterface  # noqa: F401
from .role import Role  # noqa: F401
from .snapshot import Snapshot  # noqa: F401
from .upload import Upload  # noqa: F401
from .webhook import Webhook  # noqa: F401
from .resource import Link  # noqa: F401
from .content_type import ContentType  # noqa: F401
from .content_type_field import ContentTypeField  # noqa: F401


__version__ = "2.1.1"
__author__ = "Contentful GmbH (David Litvak Bruno)"
__email__ = "david.litvak@contentful.com"



from .resource import Resource, Link


"""
contentful_management.preview_api_key
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the PreviewApiKey class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys/preview-api-key/get-a-single-preview-api-key

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class PreviewApiKey(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys/preview-api-key/get-a-single-preview-api-key
    """

    def __init__(self, item, **kwargs):
        super(PreviewApiKey, self).__init__(item, **kwargs)
        self.name = item.get('name', '')
        self.description = item.get('description', '')
        self.access_token = item.get('accessToken', '')
        self.environments = [Link(e, **kwargs) for e in item.get('environments', [])]

    def __repr__(self):
        return "<PreviewApiKey[{0}] id='{1}' access_token='{2}'>".format(
            self.name,
            self.sys.get('id', ''),
            self.access_token
        )



from .space_resource_proxy import SpaceResourceProxy
from .environments_proxy import EnvironmentsProxy

"""
contentful_management.space_environments_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpaceEnvironmentsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpaceEnvironmentsProxy(SpaceResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/
    """

    def _resource_proxy_class(self):
        return EnvironmentsProxy



from .resource import Resource
from .webhook_webhooks_call_proxy import WebhookWebhooksCallProxy
from .webhook_webhooks_health_proxy import WebhookWebhooksHealthProxy


"""
contentful_management.webhook
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Webhook class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhooks

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Webhook(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhooks
    """

    def __init__(self, item, **kwargs):
        super(Webhook, self).__init__(item, **kwargs)
        self.name = item.get('name', '')
        self.url = item.get('url', '')
        self.topics = item.get('topics', [])
        self.http_basic_username = item.get('httpBasicUsername', '')
        self.headers = item.get('headers', [])

    @classmethod
    def update_attributes_map(klass):
        """
        Defines keys and default values for non-generic attributes.
        """

        return {
            'name': '',
            'url': '',
            'topics': [],
            'http_basic_username': '',
            'headers': []
        }

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """
        Attributes for webhook creation.
        """

        result = super(Webhook, klass).create_attributes(attributes, previous_object)

        if 'topics' not in result:
            raise Exception("Topics ('topics') must be provided for this operation.")
        return result

    def calls(self):
        """
        Provides access to call overview for the given webhook.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls

        :return: :class:`WebhookWebhooksCallProxy <contentful_management.webhook_webhooks_call_proxy.WebhookWebhooksCallProxy>` object.
        :rtype: contentful.webhook_webhooks_call_proxy.WebhookWebhooksCallProxy

        Usage:

            >>> webhook_webhooks_call_proxy = webhook.calls()
            <WebhookWebhooksCallProxy space_id="cfexampleapi" webhook_id="my_webhook">
        """
        return WebhookWebhooksCallProxy(self._client, self.sys['space'].id, self.sys['id'])

    def health(self):
        """
        Provides access to health overview for the given webhook.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-health

        :return: :class:`WebhookWebhooksHealthProxy <contentful_management.webhook_webhooks_health_proxy.WebhookWebhooksHealthProxy>` object.
        :rtype: contentful.webhook_webhooks_health_proxy.WebhookWebhooksHealthProxy

        Usage:

            >>> webhook_webhooks_health_proxy = webhook.health()
            <WebhookWebhooksHealthProxy space_id="cfexampleapi" webhook_id="my_webhook">
        """
        return WebhookWebhooksHealthProxy(self._client, self.sys['space'].id, self.sys['id'])

    def to_json(self):
        """
        Returns the JSON representation of the webhook.
        """

        result = super(Webhook, self).to_json()
        result.update({
            'name': self.name,
            'url': self.url,
            'topics': self.topics,
            'httpBasicUsername': self.http_basic_username,
            'headers': self.headers
        })
        return result

    def __repr__(self):
        return "<Webhook[{0}] id='{1}'>".format(
            self.name,
            self.sys.get('id', '')
        )



from .client_proxy import ClientProxy
from .webhook import Webhook


"""
contentful_management.webhooks_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the WebhooksProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhooks

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class WebhooksProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhooks
    """

    @property
    def _resource_class(self):
        return Webhook

    def create(self, attributes=None, **kwargs):
        """
        Creates a webhook with given attributes.
        """

        return super(WebhooksProxy, self).create(resource_id=None, attributes=attributes)



from .content_type_resource_proxy import ContentTypeResourceProxy
from .snapshots_proxy import SnapshotsProxy


"""
contentful_management.content_type_snapshots_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentTypeEntriesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots/content-type-snapshots-collection

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentTypeSnapshotsProxy(ContentTypeResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots/content-type-snapshots-collection
    """

    def __init__(self, client, space_id, environment_id, content_type_id):
        self.proxy = self._resource_proxy_class()(client, space_id, environment_id, content_type_id, resource_kind='content_types')

    def _resource_proxy_class(self):
        return SnapshotsProxy

    def create(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def __repr__(self):
        return "<{0} space_id='{1}' environment_id='{2}' content_type_id='{3}'>".format(
            self.__class__.__name__,
            self.proxy.space_id,
            self.proxy.environment_id,
            self.proxy.parent_resource_id
        )



from .client_proxy import ClientProxy
from .user import User


"""
contentful_management.users_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the UsersProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/users

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class UsersProxy(ClientProxy):
    def __init__(self, client):
        super(UsersProxy, self).__init__(client, None)

    @property
    def _resource_class(self):
        return User

    def create(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def all(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def find(self, *args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def me(self):
        """
        Returns the current user information.
        """

        return self.client._get(self._url())

    def _url(self, **kwargs):
        return self._resource_class.base_url()

    def __repr__(self):
        return "<{0}>".format(
            self.__class__.__name__
        )



from .resource import Resource, EnvironmentAwareResource
from copy import deepcopy


"""
contentful_management.ui_extension
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the UIExtension class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/ui-extensions

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class UIExtension(Resource, EnvironmentAwareResource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/ui-extensions
    """

    def __init__(self, item, **kwargs):
        super(UIExtension, self).__init__(item, **kwargs)
        self.extension = deepcopy(item.get('extension', {}))

    @property
    def source(self):
        return self.extension.get('src', '') or self.extension.get('srcdoc', '')

    @source.setter
    def source(self, value):
        key = 'src' if value.startswith('http') else 'srcdoc'
        self.extension[key] = value

    @property
    def name(self):
        return self.extension.get('name', '')

    @name.setter
    def name(self, value):
        self.extension['name'] = value

    @property
    def field_types(self):
        return self.extension.get('fieldTypes', [])

    @field_types.setter
    def field_types(self, value):
        self.extension['fieldTypes'] = value

    @property
    def sidebar(self):
        return self.extension.get('sidebar', False)

    @sidebar.setter
    def sidebar(self, value):
        self.extension['sidebar'] = value

    @classmethod
    def update_attributes_map(klass):
        """
        Defines keys and default values for non-generic attributes.
        """

        return {
            'extension': {}
        }

    def to_json(self):
        """
        Returns the JSON Representation of the UI extension.
        """

        result = super(UIExtension, self).to_json()
        result.update({
            'extension': self.extension
        })
        return result

    def __repr__(self):
        return "<UIExtension[{0}] id='{1}' field_types=[{2}]>".format(
            self.name,
            self.sys.get('id', ''),
            ', '.join(
                "'{0}'".format(t)
                for t in [
                    ft['type']
                    for ft in self.field_types
                ]
            )
        )



from .space_resource_proxy import SpaceResourceProxy
from .roles_proxy import RolesProxy


"""
contentful_management.space_roles_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpaceRolesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpaceRolesProxy(SpaceResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles
    """

    def _resource_proxy_class(self):
        return RolesProxy



from .resource import Resource


"""
contentful_management.upload
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Upload class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/uploads

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Upload(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/uploads
    """

    def __init__(self, item, **kwargs):
        super(Upload, self).__init__(item, **kwargs)

    @classmethod
    def create_headers(klass, _attributes):
        """
        Headers for upload creation.
        """

        return {
            'Content-Type': 'application/octet-stream'
        }

    def __repr__(self):
        return "<Upload id='{0}'>".format(
            self.sys.get('id', '')
        )



from .client_proxy import ClientProxy
from .api_key import ApiKey


"""
contentful_management.api_keys_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ApiKeysProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ApiKeysProxy(ClientProxy):
    @property
    def _resource_class(self):
        return ApiKey

    def create(self, attributes=None, **kwargs):
        """
        Creates an API key with given attributes.
        """

        return super(ApiKeysProxy, self).create(None, attributes)



from .client_proxy import ClientProxy
from .personal_access_token import PersonalAccessToken


"""
contentful_management.personal_access_token_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the PersonalAccessTokensProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/personal-access-tokens

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class PersonalAccessTokensProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/personal-access-tokens
    """

    def __init__(self, client):
        super(PersonalAccessTokensProxy, self).__init__(client, None)

    def __repr__(self):
        return "<PersonalAccessTokensProxy>"

    def _url(self, resource_id='', **kwargs):
        return self._resource_class.base_url(resource_id=resource_id, **kwargs)

    @property
    def _resource_class(self):
        return PersonalAccessToken

    def create(self, attributes):
        return super(PersonalAccessTokensProxy, self).create(None, attributes)

    def delete(self, token_id, *args, **kwargs):
        """
        Revokes a personal access token.
        """

        return self.client._put(
            "{0}/revoked".format(
                self._url(token_id)
            ),
            None,
            *args,
            **kwargs
        )

    def revoke(self, token_id, *args, **kwargs):
        """
        Revokes a personal access token.
        """

        return self.delete(token_id, *args, **kwargs)



from .client_proxy import ClientProxy
from .locale import Locale


"""
contentful_management.locales_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the LocalesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class LocalesProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales
    """

    @property
    def _resource_class(self):
        return Locale

    def create(self, attributes=None, environment_id=None, **kwargs):
        """
        Creates a locale with given attributes.
        """

        return super(LocalesProxy, self).create(None, **kwargs)



from .environment_resource_proxy import EnvironmentResourceProxy
from .content_types_proxy import ContentTypesProxy


"""
contentful_management.environment_content_types_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EnvironmentContentTypesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EnvironmentContentTypesProxy(EnvironmentResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types
    """

    def _resource_proxy_class(self):
        return ContentTypesProxy



from .client_proxy import ClientProxy
from .snapshot import Snapshot


"""
contentful_management.snapshots_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SnapshotsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SnapshotsProxy(ClientProxy):
    def __init__(self, client, space_id, environment_id, parent_resource_id, resource_kind='entries'):
        super(SnapshotsProxy, self).__init__(client, space_id, environment_id=environment_id)
        self.parent_resource_id = parent_resource_id
        self.resource_kind = resource_kind

    @property
    def _resource_class(self):
        return Snapshot

    def create(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def delete(*args, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not Supported")

    def _url(self, resource_id='', **kwargs):
        return self._resource_class.base_url(
            self.space_id,
            self.parent_resource_id,
            environment_id=self.environment_id,
            resource_url=self.resource_kind,
            resource_id=resource_id
        )

    def __repr__(self):
        return "<{0}[{1}] space_id='{2}' environment_id='{3}' parent_resource_id='{4}'>".format(
            self.__class__.__name__,
            self.resource_kind,
            self.space_id,
            self.environment_id,
            self.parent_resource_id
        )



from .space_resource_proxy import SpaceResourceProxy
from .api_keys_proxy import ApiKeysProxy


"""
contentful_management.space_api_keys_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the SpaceApiKeysProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class SpaceApiKeysProxy(SpaceResourceProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys
    """

    def _resource_proxy_class(self):
        return ApiKeysProxy



from .resource import Resource, Link


"""
contentful_management.api_key
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ApiKey class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ApiKey(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys
    """

    def __init__(self, item, **kwargs):
        super(ApiKey, self).__init__(item, **kwargs)
        self.name = item.get('name', '')
        self.description = item.get('description', '')
        self.access_token = item.get('accessToken', '')
        self.policies = item.get('policies', [])
        self.environments = [Link(e, **kwargs) for e in item.get('environments', [])]

        preview_api_key = item.get('preview_api_key', None)
        self._preview_api_key = Link(preview_api_key, **kwargs) if preview_api_key is not None else None

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """
        Attributes for resource creation.
        """

        return {
            'name': attributes.get(
                'name',
                previous_object.name if previous_object is not None else ''
            ),
            'description': attributes.get(
                'description',
                previous_object.description if previous_object is not None else ''
            ),
            'environments': attributes.get(
                'environments',
                [e.to_json() for e in previous_object.environments] if previous_object is not None else []  # Will default to master if empty
            )
        }

    @classmethod
    def update_attributes_map(klass):
        """
        Defines keys and default values for non-generic attributes.
        """

        return {
            'name': '',
            'description': '',
            'access_token': '',
            'environments': []
        }

    def preview_api_key(self):
        return self._preview_api_key.resolve(self.space.id)

    def to_json(self):
        """
        Returns the JSON representation of the API key.
        """

        result = super(ApiKey, self).to_json()
        result.update({
            'name': self.name,
            'description': self.description,
            'accessToken': self.access_token,
            'environments': [e.to_json() for e in self.environments]
        })
        return result

    def __repr__(self):
        return "<ApiKey[{0}] id='{1}' access_token='{2}'>".format(
            self.name,
            self.sys.get('id', ''),
            self.access_token
        )



"""
contentful_management.array
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Array class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/introduction/collection-resources-and-pagination

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Array(object):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/introduction/collection-resources-and-pagination
    """

    def __init__(self, json, items):
        self.items = items
        self.skip = json.get('skip', 0)
        self.limit = json.get('limit', 100)
        self.total = json.get('total', 0)

    def __iter__(self):
        for item in self.items:
            yield item

    def __getitem__(self, index):
        return self.items[index]

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return "<Array size='{0}' total='{1}' limit='{2}' skip='{3}'>".format(
            len(self),
            self.total,
            self.limit,
            self.skip
        )



from .resource import Resource


"""
contentful_management.webhook_health
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the WebhookHealth class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-health

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class WebhookHealth(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-health
    """

    def __init__(self, item, **kwargs):
        super(WebhookHealth, self).__init__(item, **kwargs)
        self.calls = item.get('calls', {})
        self.total = self.calls.get('total', 0)
        self.healthy = self.calls.get('healthy', 0)

    @classmethod
    def base_url(klass, space_id, webhook_id, **kwargs):
        """
        Returns the URI for the webhook health.
        """

        return "spaces/{0}/webhooks/{1}/health".format(
            space_id,
            webhook_id
        )

    @property
    def webhook_id(self):
        return self.sys['created_by'].sys['id']

    def __repr__(self):
        return "<WebhookHealth[{0}] total={1} healthy={2}>".format(
            self.webhook_id,
            self.total,
            self.healthy
        )



from .resource import Resource
from .environment_content_types_proxy import EnvironmentContentTypesProxy
from .environment_entries_proxy import EnvironmentEntriesProxy
from .environment_assets_proxy import EnvironmentAssetsProxy
from .environment_locales_proxy import EnvironmentLocalesProxy
from .environment_ui_extensions_proxy import EnvironmentUIExtensionsProxy

"""
contentful_management.environment
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Environment class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Environment(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales
    """

    def __init__(self, item, **kwargs):
        super(Environment, self).__init__(item, **kwargs)
        self.name = item.get('name', '')

    @classmethod
    def base_url(klass, space_id, resource_id=None, **kwargs):
        """
        Returns the URI for the environment.
        """

        return "spaces/{0}/environments/{1}".format(space_id, resource_id)

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """
        Attributes for environment creation.
        """

        return super(Environment, klass).create_attributes(attributes, previous_object)

    def to_json(self):
        """
        Returns the JSON representation of the environment.
        """

        result = super(Environment, self).to_json()
        result.update({
            'name': self.name
        })

        return result

    def content_types(self):
        """
        Provides access to content type management methods for content types of an environment.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types

        :return: :class:`EnvironmentContentTypesProxy <contentful_management.space_content_types_proxy.EnvironmentContentTypesProxy>` object.
        :rtype: contentful.space_content_types_proxy.EnvironmentContentTypesProxy

        Usage:

            >>> space_content_types_proxy = environment.content_types()
            <EnvironmentContentTypesProxy space_id="cfexampleapi" environment_id="master">
        """

        return EnvironmentContentTypesProxy(self._client, self.space.id, self.id)

    def entries(self):
        """
        Provides access to entry management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries

        :return: :class:`EnvironmentEntriesProxy <contentful_management.environment_entries_proxy.EnvironmentEntriesProxy>` object.
        :rtype: contentful.environment_entries_proxy.EnvironmentEntriesProxy

        Usage:

            >>> environment_entries_proxy = environment.entries()
            <EnvironmentEntriesProxy space_id="cfexampleapi" environment_id="master">
        """

        return EnvironmentEntriesProxy(self._client, self.space.id, self.id)

    def assets(self):
        """
        Provides access to asset management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets

        :return: :class:`EnvironmentAssetsProxy <contentful_management.environment_assets_proxy.EnvironmentAssetsProxy>` object.
        :rtype: contentful.environment_assets_proxy.EnvironmentAssetsProxy

        Usage:

            >>> environment_assets_proxy = environment.assets()
            <EnvironmentAssetsProxy space_id="cfexampleapi" environment_id="master">
        """

        return EnvironmentAssetsProxy(self._client, self.space.id, self.id)

    def locales(self):
        """
        Provides access to locale management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales

        :return: :class:`EnvironmentLocalesProxy <contentful_management.environment_locales_proxy.EnvironmentLocalesProxy>` object.
        :rtype: contentful.environment_locales_proxy.EnvironmentLocalesProxy

        Usage:

            >>> environment_locales_proxy = environment.locales()
            <EnvironmentLocalesProxy space_id="cfexampleapi" environment_id="master">
        """

        return EnvironmentLocalesProxy(self._client, self.space.id, self.id)

    def ui_extensions(self):
        """
        Provides access to UI extensions management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/ui-extensions

        :return: :class:`EnvironmentUIExtensionsProxy <contentful_management.ui_extensions_proxy.EnvironmentUIExtensionsProxy>` object.
        :rtype: contentful.ui_extensions_proxy.EnvironmentUIExtensionsProxy

        Usage:

            >>> ui_extensions_proxy = environment.ui_extensions()
            <EnvironmentUIExtensionsProxy space_id="cfexampleapi" environment_id="master">
        """

        return EnvironmentUIExtensionsProxy(self._client, self.space.id, self.id)

    def __repr__(self):
        return "<Environment[{0}] id='{1}'>".format(
            self.name,
            self.id
        )



import requests
import platform
from re import sub

from .resource_builder import ResourceBuilder
from .errors import get_error, RateLimitExceededError
from .utils import ConfigurationException, retry_request

from .users_proxy import UsersProxy
from .roles_proxy import RolesProxy
from .assets_proxy import AssetsProxy
from .spaces_proxy import SpacesProxy
from .entries_proxy import EntriesProxy
from .locales_proxy import LocalesProxy
from .uploads_proxy import UploadsProxy
from .api_keys_proxy import ApiKeysProxy
from .webhooks_proxy import WebhooksProxy
from .snapshots_proxy import SnapshotsProxy
from .environments_proxy import EnvironmentsProxy
from .webhooks_call_proxy import WebhooksCallProxy
from .ui_extensions_proxy import UIExtensionsProxy
from .content_types_proxy import ContentTypesProxy
from .organizations_proxy import OrganizationsProxy
from .webhooks_health_proxy import WebhooksHealthProxy
from .preview_api_keys_proxy import PreviewApiKeysProxy
from .space_memberships_proxy import SpaceMembershipsProxy
from .editor_interfaces_proxy import EditorInterfacesProxy
from .personal_access_tokens_proxy import PersonalAccessTokensProxy

try:
    import multijson as json
except ImportError:
    import json


"""
contentful_management.client
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Contentful Management API Client,
allowing interaction with every method present in it.

Complete API documentation: https://www.contentful.com/developers/docs/references/content-management-api

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Client(object):
    """Constructs the API client.

    :param access_token: API access token.
    :param api_url: (optional) URL of the Contentful API,
        defaults to Management API.
    :param uploads_api_url: (optional) URL of the Contentful upload API,
        defaults to Upload API.
    :param api_version: (optional) Target version of the Contentful API.
    :param default_locale: (optional) Default locale for your spaces,
        defaults to 'en-US'.
    :param https: (optional) Boolean determining wether to use https
        or http, defaults to True.
    :param raw_mode: (optional) Boolean determining wether to process the
        response or return it raw after each API call, defaults to True.
    :param gzip_encoded: (optional) Boolean determining wether to accept
        gzip encoded results, defaults to True.
    :param raise_errors: (optional) Boolean determining wether to raise
        an exception on requests that aren't successful, defaults to True.
    :param proxy_host: (optional) URL for Proxy, defaults to None.
    :param proxy_port: (optional) Port for Proxy, defaults to None.
    :param proxy_username: (optional) Username for Proxy, defaults to None.
    :param proxy_password: (optional) Password for Proxy, defaults to None.
    :param max_rate_limit_retries: (optional) Maximum amount of retries
        after RateLimitError, defaults to 1.
    :param max_rate_limit_wait: (optional) Timeout (in seconds) for waiting
        for retry after RateLimitError, defaults to 60.
    :param application_name: (optional) User application name, defaults to None.
    :param application_version: (optional) User application version, defaults to None.
    :param integration_name: (optional) Integration name, defaults to None.
    :param integration_version: (optional) Integration version, defaults to None.
    :return: :class:`Client <Client>` object.
    :rtype: contentful.Client

    Usage:

        >>> import contentful_management
        >>> client = contentful_management.Client('YOUR_MANAGEMENT_TOKEN')
        <contentful_management.Client access_token="YOUR_MANAGEMENT_TOKEN" default_locale="en-US">
    """

    def __init__(
            self,
            access_token,
            api_url='api.contentful.com',
            uploads_api_url='upload.contentful.com',
            api_version=1,
            default_locale='en-US',
            https=True,
            raw_mode=False,
            gzip_encoded=True,
            raise_errors=True,
            proxy_host=None,
            proxy_port=None,
            proxy_username=None,
            proxy_password=None,
            max_rate_limit_retries=1,
            max_rate_limit_wait=60,
            application_name=None,
            application_version=None,
            integration_name=None,
            integration_version=None,
            additional_headers=None):
        self.access_token = access_token
        self.api_url = api_url
        self.uploads_api_url = uploads_api_url
        self.api_version = api_version
        self.default_locale = default_locale
        self.https = https
        self.raw_mode = raw_mode
        self.gzip_encoded = gzip_encoded
        self.raise_errors = raise_errors
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.proxy_username = proxy_username
        self.proxy_password = proxy_password
        self.max_rate_limit_retries = max_rate_limit_retries
        self.max_rate_limit_wait = max_rate_limit_wait
        self.application_name = application_name
        self.application_version = application_version
        self.integration_name = integration_name
        self.integration_version = integration_version
        self.additional_headers = additional_headers or {}

        self._validate_configuration()

    def spaces(self):
        """
        Provides access to space management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/spaces

        :return: :class:`SpacesProxy <contentful_management.spaces_proxy.SpacesProxy>` object.
        :rtype: contentful.spaces_proxy.SpacesProxy

        Usage:

            >>> spaces_proxy = client.spaces()
            <SpacesProxy>
        """

        return SpacesProxy(self)

    def memberships(self, space_id):
        """
        Provides access to space membership management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/space-memberships

        :return: :class:`SpaceMembershipsProxy <contentful_management.space_memberships_proxy.SpaceMembershipsProxy>` object.
        :rtype: contentful.space_memberships_proxy.SpaceMembershipsProxy

        Usage:

            >>> space_memberships_proxy = client.memberships('cfexampleapi')
            <SpaceMembershipsProxy space_id="cfexampleapi">
        """

        return SpaceMembershipsProxy(self, space_id)

    def organizations(self):
        """
        Provides access to organization management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/organizations

        :return: :class:`OrganizationsProxy <contentful_management.organizations_proxy.OrganizationsProxy>` object.
        :rtype: contentful.organizations_proxy.OrganizationsProxy

        Usage:

            >>> organizations_proxy = client.organizations()
            <OrganizationsProxy>
        """

        return OrganizationsProxy(self)

    def users(self):
        """
        Provides access to user management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/users

        :return: :class:`UsersProxy <contentful_management.users_proxy.UsersProxy>` object.
        :rtype: contentful.users_proxy.UsersProxy

        Usage:

            >>> users_proxy = client.users()
            <UsersProxy>
        """

        return UsersProxy(self)

    def content_types(self, space_id, environment_id):
        """
        Provides access to content type management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types

        :return: :class:`ContentTypesProxy <contentful_management.content_types_proxy.ContentTypesProxy>` object.
        :rtype: contentful.content_types_proxy.ContentTypesProxy

        Usage:

            >>> content_types_proxy = client.content_types('cfexampleapi', 'master')
            <ContentTypesProxy space_id="cfexampleapi" environment_id="master">
        """

        return ContentTypesProxy(self, space_id, environment_id)

    def entries(self, space_id, environment_id):
        """
        Provides access to entry management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/entries

        :return: :class:`EntriesProxy <contentful_management.entries_proxy.EntriesProxy>` object.
        :rtype: contentful.entries_proxy.EntriesProxy

        Usage:

            >>> entries_proxy = client.entries('cfexampleapi', 'master')
            <EntriesProxy space_id="cfexampleapi" environment_id="master">
        """

        return EntriesProxy(self, space_id, environment_id)

    def assets(self, space_id, environment_id):
        """
        Provides access to asset management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/assets

        :return: :class:`AssetsProxy <contentful_management.assets_proxy.AssetsProxy>` object.
        :rtype: contentful.assets_proxy.AssetsProxy

        Usage:

            >>> assets_proxy = client.assets('cfexampleapi', 'master')
            <AssetsProxy space_id="cfexampleapi" environment_id="master">
        """

        return AssetsProxy(self, space_id, environment_id)

    def locales(self, space_id, environment_id):
        """
        Provides access to locale management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/locales

        :return: :class:`LocalesProxy <contentful_management.locales_proxy.LocalesProxy>` object.
        :rtype: contentful.locales_proxy.LocalesProxy

        Usage:

            >>> locales_proxy = client.locales('cfexampleapi', 'master')
            <LocalesProxy space_id="cfexampleapi" environment_id="master">
        """

        return LocalesProxy(self, space_id, environment_id)

    def webhooks(self, space_id):
        """
        Provides access to webhook management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhooks

        :return: :class:`WebhooksProxy <contentful_management.locales_proxy.WebhooksProxy>` object.
        :rtype: contentful.webhooks_proxy.WebhooksProxy

        Usage:

            >>> webhooks_proxy = client.webhooks('cfexampleapi')
            <WebhooksProxy space_id="cfexampleapi">
        """

        return WebhooksProxy(self, space_id)

    def webhook_calls(self, space_id, webhook_id):
        """
        Provides access to webhook call information.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls

        :return: :class:`WebhooksProxy <contentful_management.webhooks_call_proxy.WebhooksCallProxy>` object.
        :rtype: contentful.webhooks_call_proxy.WebhooksProxy

        Usage:

            >>> webhooks_call_proxy = client.webhook_calls('cfexampleapi', 'my_webhook')
            <WebhooksCallProxy space_id="cfexampleapi" webhook_id="my_webhook">
        """

        return WebhooksCallProxy(self, space_id, webhook_id)

    def webhook_health(self, space_id, webhook_id):
        """
        Provides access to webhook health information.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhook-calls/webhook-health

        :return: :class:`WebhooksHealthProxy <contentful_management.webhooks_health_proxy.WebhooksHealthProxy>` object.
        :rtype: contentful.webhooks_health_proxy.WebhooksHealthProxy

        Usage:

            >>> webhooks_health_proxy = client.webhook_calls('cfexampleapi', 'my_webhook')
            <WebhooksHealthProxy space_id="cfexampleapi" webhook_id="my_webhook">
        """

        return WebhooksHealthProxy(self, space_id, webhook_id)

    def api_keys(self, space_id):
        """
        Provides access to API key management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys

        :return: :class:`ApiKeysProxy <contentful_management.api_keys_proxy.ApiKeysProxy>` object.
        :rtype: contentful.api_keys_proxy.ApiKeysProxy

        Usage:

            >>> api_keys_proxy = client.api_keys('cfexampleapi')
            <ApiKeysProxy space_id="cfexampleapi">
        """

        return ApiKeysProxy(self, space_id)

    def preview_api_keys(self, space_id):
        """
        Provides access to Preview API key management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys/preview-api-key/get-a-single-preview-api-key

        :return: :class:`PreviewApiKeysProxy <contentful_management.preview_api_keys_proxy.PreviewApiKeysProxy>` object.
        :rtype: contentful.preview_api_keys_proxy.PreviewApiKeysProxy

        Usage:

            >>> preview_api_keys_proxy = client.preview_api_keys('cfexampleapi')
            <PreviewApiKeysProxy space_id="cfexampleapi">
        """

        return PreviewApiKeysProxy(self, space_id)

    def personal_access_tokens(self):
        """
        Provides access to personal access token management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/personal-access-tokens

        :return: :class:`PersonalAccessTokensProxy <contentful_management.personal_access_tokens_proxy.PersonalAccessTokensProxy>` object.
        :rtype: contentful.personal_access_tokens_proxy.PersonalAccessTokensProxy

        Usage:

            >>> personal_access_tokens_proxy = client.personal_access_tokens()
            <PersonalAccessTokensProxy>
        """

        return PersonalAccessTokensProxy(self)

    def roles(self, space_id):
        """
        Provides access to role management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles

        :return: :class:`RolesProxy <contentful_management.roles_proxy.RolesProxy>` object.
        :rtype: contentful.roles_proxy.RolesProxy

        Usage:

            >>> roles_proxy = client.roles('cfexampleapi')
            <RolesProxy space_id="cfexampleapi">
        """

        return RolesProxy(self, space_id)

    def ui_extensions(self, space_id, environment_id):
        """
        Provides access to UI extensions management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/ui-extensions

        :return: :class:`UIExtensionsProxy <contentful_management.ui_extensions_proxy.UIExtensionsProxy>` object.
        :rtype: contentful.ui_extensions_proxy.UIExtensionsProxy

        Usage:

            >>> ui_extensions_proxy = client.ui_extensions('cfexampleapi', 'master')
            <UIExtensionsProxy space_id="cfexampleapi" environment_id="master">
        """

        return UIExtensionsProxy(self, space_id, environment_id)

    def editor_interfaces(self, space_id, environment_id, content_type_id):
        """
        Provides access to editor interfaces management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface

        :return: :class:`EditorInterfacesProxy <contentful_management.editor_interfaces_proxy.EditorInterfacesProxy>` object.
        :rtype: contentful.editor_interfaces_proxy.EditorInterfacesProxy

        Usage:

            >>> editor_interfaces_proxy = client.editor_interfaces('cfexampleapi', 'master', 'cat')
            <EditorInterfacesProxy space_id="cfexampleapi" environment_id="master" content_type_id="cat">
        """

        return EditorInterfacesProxy(self, space_id, environment_id, content_type_id)

    def snapshots(self, space_id, environment_id, resource_id, resource_kind='entries'):
        """
        Provides access to snapshot management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

        :return: :class:`SnapshotsProxy <contentful_management.snapshots_proxy.SnapshotsProxy>` object.
        :rtype: contentful.snapshots_proxy.SnapshotsProxy

        Usage:

            >>> entry_snapshots_proxy = client.snapshots('cfexampleapi', 'master', 'nyancat')
            <SnapshotsProxy[entries] space_id="cfexampleapi" environment_id="master" parent_resource_id="nyancat">

            >>> content_type_snapshots_proxy = client.snapshots('cfexampleapi', 'master', 'cat', 'content_types')
            <SnapshotsProxy[content_types] space_id="cfexampleapi" environment_id="master" parent_resource_id="cat">
        """

        return SnapshotsProxy(self, space_id, environment_id, resource_id, resource_kind)

    def entry_snapshots(self, space_id, environment_id, entry_id):
        """
        Provides access to entry snapshot management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

        :return: :class:`SnapshotsProxy <contentful_management.snapshots_proxy.SnapshotsProxy>` object.
        :rtype: contentful.snapshots_proxy.SnapshotsProxy

        Usage:

            >>> entry_snapshots_proxy = client.entry_snapshots('cfexampleapi', 'master', 'nyancat')
            <SnapshotsProxy[entries] space_id="cfexampleapi" environment_id="master" parent_resource_id="nyancat">
        """

        return SnapshotsProxy(self, space_id, environment_id, entry_id, 'entries')

    def content_type_snapshots(self, space_id, environment_id, content_type_id):
        """
        Provides access to content type snapshot management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/snapshots

        :return: :class:`SnapshotsProxy <contentful_management.snapshots_proxy.SnapshotsProxy>` object.
        :rtype: contentful.snapshots_proxy.SnapshotsProxy

        Usage:

            >>> content_type_snapshots_proxy = client.content_type_snapshots('cfexampleapi', 'master', 'cat')
            <SnapshotsProxy[content_types] space_id="cfexampleapi" environment_id="master" parent_resource_id="cat">
        """

        return SnapshotsProxy(self, space_id, environment_id, content_type_id, 'content_types')

    def uploads(self, space_id):
        """
        Provides access to upload management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/uploads

        :return: :class:`UploadsProxy <contentful_management.uploads_proxy.UploadsProxy>` object.
        :rtype: contentful.uploads_proxy.UploadsProxy

        Usage:

            >>> uploads_proxy = client.uploads('cfexampleapi')
            <UploadsProxy space_id="cfexampleapi">
        """

        return UploadsProxy(self, space_id)

    def environments(self, space_id):
        """
        Provides access to environment management methods.

        API reference: TBD

        :return: :class:`TBD` object.
        :rtype: contentful.environments_proxy.EnvironmentsProxy

        Usage:

            >>> environments_proxy = client.environments('cfexampleapi')
            <EnvironmentsProxy space_id="cfexampleapi">
        """

        return EnvironmentsProxy(self, space_id)

    def _validate_configuration(self):
        """
        Validates that required parameters are present.
        """

        if not self.access_token:
            raise ConfigurationException(
                'You will need to initialize a client with an Access Token'
            )
        if not self.api_url:
            raise ConfigurationException(
                'The client configuration needs to contain an API URL'
            )
        if not self.default_locale:
            raise ConfigurationException(
                'The client configuration needs to contain a Default Locale'
            )
        if not self.api_version or self.api_version < 1:
            raise ConfigurationException(
                'The API Version must be a positive number'
            )

    def _contentful_user_agent(self):
        """
        Sets the X-Contentful-User-Agent header.
        """
        header = {}
        from . import __version__
        header['sdk'] = {
            'name': 'contentful-management.py',
            'version': __version__
        }
        header['app'] = {
            'name': self.application_name,
            'version': self.application_version
        }
        header['integration'] = {
            'name': self.integration_name,
            'version': self.integration_version
        }
        header['platform'] = {
            'name': 'python',
            'version': platform.python_version()
        }

        os_name = platform.system()
        if os_name == 'Darwin':
            os_name = 'macOS'
        elif not os_name or os_name == 'Java':
            os_name = None
        elif os_name and os_name not in ['macOS', 'Windows']:
            os_name = 'Linux'
        header['os'] = {
            'name': os_name,
            'version': platform.release()
        }

        def format_header(key, values):
            header = "{0} {1}".format(key, values['name'])
            if values['version'] is not None:
                header = "{0}/{1}".format(header, values['version'])
            return "{0};".format(header)

        result = []
        for k, values in header.items():
            if not values['name']:
                continue
            result.append(format_header(k, values))

        return ' '.join(result)

    def _request_headers(self):
        """
        Sets the default request headers.
        """

        headers = {
            'X-Contentful-User-Agent': self._contentful_user_agent(),
            'Content-Type':
                'application/vnd.contentful.management.v{0}+json'.format(
                    self.api_version
                ),
            'Authorization': 'Bearer {0}'.format(self.access_token)
        }

        if self.gzip_encoded:
            headers['Accept-Encoding'] = 'gzip'
        else:
            headers['Accept-Encoding'] = 'identity'

        return headers

    def _url(self, url, file_upload=False):
        """
        Creates the request URL.
        """

        host = self.api_url
        if file_upload:
            host = self.uploads_api_url

        protocol = 'https' if self.https else 'http'
        if url.endswith('/'):
            url = url[:-1]
        return '{0}://{1}/{2}'.format(
            protocol,
            host,
            url
        )

    def _normalize_query(self, query):
        """
        Converts arrays in the query to comma
        separated lists for proper API handling.
        """

        for k, v in query.items():
            if isinstance(v, list):
                query[k] = ','.join(v)

    def _http_request(self, method, url, request_kwargs=None):
        """
        Performs the requested HTTP request.
        """

        kwargs = request_kwargs if request_kwargs is not None else {}

        headers = self._request_headers()
        headers.update(self.additional_headers)
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
        kwargs['headers'] = headers

        if self._has_proxy():
            kwargs['proxies'] = self._proxy_parameters()

        request_url = self._url(
            url,
            file_upload=kwargs.pop('file_upload', False)
        )

        request_method = getattr(requests, method)
        response = request_method(request_url, **kwargs)

        if response.status_code == 429:
            raise RateLimitExceededError(response)

        return response

    def _http_get(self, url, query, **kwargs):
        """
        Performs the HTTP GET request.
        """

        self._normalize_query(query)

        kwargs.update({'params': query})

        return self._http_request('get', url, kwargs)

    def _http_post(self, url, data, **kwargs):
        """
        Performs the HTTP POST request.
        """

        if not kwargs.get('file_upload', False):
            data = json.dumps(data)

        kwargs.update({'data': data})

        return self._http_request('post', url, kwargs)

    def _http_put(self, url, data, **kwargs):
        """
        Performs the HTTP PUT request.
        """

        kwargs.update({'data': json.dumps(data)})

        return self._http_request('put', url, kwargs)

    def _http_delete(self, url, _data, **kwargs):
        """
        Performs the HTTP DELETE request.
        """

        return self._http_request('delete', url, kwargs)

    def _request(self, method, url, query_or_data=None, **kwargs):
        """
        Wrapper for the HTTP requests,
        rate limit backoff is handled here,
        responses are processed with ResourceBuilder.
        """

        if query_or_data is None:
            query_or_data = {}

        request_method = getattr(self, '_http_{0}'.format(method))
        response = retry_request(self)(request_method)(url, query_or_data, **kwargs)

        if self.raw_mode:
            return response

        if response.status_code >= 300:
            error = get_error(response)
            if self.raise_errors:
                raise error
            return error

        # Return response object on NoContent
        if response.status_code == 204 or not response.text:
            return response

        return ResourceBuilder(
            self,
            self.default_locale,
            response.json()
        ).build()

    def _get(self, url, query=None, **kwargs):
        """
        Wrapper for the HTTP GET request.
        """

        return self._request('get', url, query, **kwargs)

    def _post(self, url, attributes=None, **kwargs):
        """
        Wrapper for the HTTP POST request.
        """

        return self._request('post', url, attributes, **kwargs)

    def _put(self, url, attributes=None, **kwargs):
        """
        Wrapper for the HTTP PUT request.
        """

        return self._request('put', url, attributes, **kwargs)

    def _delete(self, url, **kwargs):
        """
        Wrapper for the HTTP DELETE request.
        """

        response = retry_request(self)(self._http_delete)(url, **kwargs)

        if self.raw_mode:
            return response

        if response.status_code >= 300:
            error = get_error(response)
            if self.raise_errors:
                raise error
            return error

        return response

    def _has_proxy(self):
        """
        Checks if a proxy was set.
        """

        return self.proxy_host

    def _proxy_parameters(self):
        """
        Builds proxy parameters dict from
        client options.
        """

        proxy_protocol = ''
        if self.proxy_host.startswith('https'):
            proxy_protocol = 'https'
        else:
            proxy_protocol = 'http'

        proxy = '{0}://'.format(proxy_protocol)
        if self.proxy_username and self.proxy_password:
            proxy += '{0}:{1}@'.format(self.proxy_username, self.proxy_password)

        proxy += sub(r'https?(://)?', '', self.proxy_host)

        if self.proxy_port:
            proxy += ':{0}'.format(self.proxy_port)

        return {
            'http': proxy,
            'https': proxy
        }

    def __repr__(self):
        return "<contentful_management.Client access_token='{0}' default_locale='{1}'>".format(  # noqa: E501
            self.access_token,
            self.default_locale
        )



from .client_proxy import ClientProxy
from .content_type import ContentType


"""
contentful_management.content_types_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ContentTypesProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/editor-interface

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ContentTypesProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types/content-type
    """

    @property
    def _resource_class(self):
        return ContentType

    def all_published(self):
        """
        Gets all the published content types for a space.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/content-types/activated-content-type-collection
        """

        return self.client._get(
            self._url(public=True)
        )

    def _url(self, resource_id='', **kwargs):
        return self._resource_class.base_url(self.space_id, resource_id=resource_id, environment_id=self.environment_id, **kwargs)



from .client_proxy import ClientProxy
from .environment import Environment

"""
contentful_management.environment_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the EnvironmentsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/spaces

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class EnvironmentsProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/
    """

    def __init__(self, client, space_id):
        super(EnvironmentsProxy, self).__init__(client, space_id)

    @property
    def _resource_class(self):
        return Environment

    def all(self, query=None, **kwargs):
        """
        Gets all environments.
        """

        return super(EnvironmentsProxy, self).all(query=query)

    def find(self, environment_id, query=None, **kwargs):
        """
        Gets an environment by ID.
        """

        return super(EnvironmentsProxy, self).find(environment_id, query=query)

    def create(self, environment_id=None, attributes=None):
        """
        Creates an environment with a given ID (optional) and attributes.
        """

        return super(EnvironmentsProxy, self).create(environment_id, attributes)

    def delete(self, environment_id):
        """
        Deletes an environment by ID.
        """

        return super(EnvironmentsProxy, self).delete(environment_id)



from .resource import Resource


"""
contentful_management.organization
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Organization class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/organizations

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Organization(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/organizations
    """

    def __init__(self, item, **kwargs):
        super(Organization, self).__init__(item, **kwargs)
        self.name = item.get('name', '')

    @classmethod
    def base_url(klass, *args, **kwargs):
        """
        Returns the URI for the organization.
        """

        return "organizations"

    def __repr__(self):
        return "<Organization[{0}] id='{1}'>".format(
            self.name,
            self.id
        )



from .role import Role
from .user import User
from .entry import Entry
from .array import Array
from .asset import Asset
from .space import Space
from .locale import Locale
from .upload import Upload
from .api_key import ApiKey
from .webhook import Webhook
from .snapshot import Snapshot
from .environment import Environment
from .ui_extension import UIExtension
from .content_type import ContentType
from .webhook_call import WebhookCall
from .organization import Organization
from .webhook_health import WebhookHealth
from .preview_api_key import PreviewApiKey
from .editor_interface import EditorInterface
from .space_membership import SpaceMembership
from .personal_access_token import PersonalAccessToken


"""
contentful_management.resource_builder
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the ResourceBuilder class.

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class ResourceBuilder(object):
    """
    Creates objects of the proper resource type.
    """

    def __init__(self, client, default_locale, json):
        self.client = client
        self.default_locale = default_locale
        self.json = json

    def build(self):
        """
        Creates the objects from the JSON response.
        """

        if self.json['sys']['type'] == 'Array':
            return self._build_array()
        return self._build_item(self.json)

    def _build_array(self):
        return Array(self.json, [self._build_item(item) for item in self.json['items']])

    def _build_item(self, item):
        buildables = {
            'Entry': Entry,
            'Asset': Asset,
            'ContentType': ContentType,
            'Space': Space,
            'SpaceMembership': SpaceMembership,
            'Organization': Organization,
            'ApiKey': ApiKey,
            'PreviewApiKey': PreviewApiKey,
            'PersonalAccessToken': PersonalAccessToken,
            'Locale': Locale,
            'EditorInterface': EditorInterface,
            'WebhookDefinition': Webhook,
            'WebhookCallOverview': WebhookCall,
            'WebhookCallDetails': WebhookCall,
            'Webhook': WebhookHealth,
            'Role': Role,
            'Extension': UIExtension,
            'Snapshot': Snapshot,
            'Upload': Upload,
            'User': User,
            'Environment': Environment
        }

        item_type = item['sys']['type']

        if item_type in buildables:
            return buildables[item_type](
                item,
                default_locale=self.default_locale,
                client=self.client
            )
        raise Exception("Resource not buildable")



from .client_proxy import ClientProxy
from .organization import Organization


"""
contentful_management.organizations_proxy
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the OrganizationsProxy class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/organizations

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class OrganizationsProxy(ClientProxy):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/organizations
    """

    def __init__(self, client):
        super(OrganizationsProxy, self).__init__(client, None)

    def __repr__(self):
        return "<OrganizationsProxy>"

    @property
    def _resource_class(self):
        return Organization

    def all(self, query=None, **kwargs):
        """
        Gets all organizations.
        """

        return super(OrganizationsProxy, self).all(query=query)

    def find(self, organization_id, query=None, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not supported")

    def create(self, attributes=None, **kwargs):
        """
        Not supported.
        """

        raise Exception("Not supported")

    def delete(self, organization_id):
        """
        Not supported.
        """

        raise Exception("Not supported")



from .resource import Resource
from .space_roles_proxy import SpaceRolesProxy
from .space_uploads_proxy import SpaceUploadsProxy
from .space_api_keys_proxy import SpaceApiKeysProxy
from .space_webhooks_proxy import SpaceWebhooksProxy
from .space_environments_proxy import SpaceEnvironmentsProxy
from .space_preview_api_keys_proxy import SpacePreviewApiKeysProxy


"""
contentful_management.space
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the Space class.

API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/spaces

:copyright: (c) 2018 by Contentful GmbH.
:license: MIT, see LICENSE for more details.
"""


class Space(Resource):
    """
    API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/spaces
    """

    def __init__(self, item, **kwargs):
        super(Space, self).__init__(item, **kwargs)
        self.name = item.get('name', '')

    @classmethod
    def base_url(klass, space_id=None, **kwargs):
        """
        Returns the URI for the space.
        """

        if space_id is None:
            space_id = ''
        return "spaces/{0}".format(space_id)

    @classmethod
    def create_headers(klass, attributes):
        """
        Headers for space creation.
        """

        if 'organization_id' in attributes:
            return {'x-contentful-organization': attributes['organization_id']}
        return {}

    @classmethod
    def create_attributes(klass, attributes, previous_object=None):
        """Attributes for space creation."""

        if previous_object is not None:
            return {'name': attributes.get('name', previous_object.name)}
        return {
            'name': attributes.get('name', ''),
            'defaultLocale': attributes['default_locale']
        }

    @classmethod
    def update_attributes_map(klass):
        """
        Defines keys and default values for non-generic attributes.
        """

        return {
            'name': '',
            'default_locale': ''
        }

    def update(self, attributes=None):
        """Updates the space with attributes."""

        if attributes is None:
            attributes = {}

        headers = self.__class__.create_headers(attributes)
        headers.update(self._update_headers())

        result = self._client._put(
            self.__class__.base_url(
                self.sys['id']
            ),
            self.__class__.create_attributes(attributes, self),
            headers=headers
        )

        self._update_from_resource(result)

        return self

    def reload(self):
        """
        Reloads the space.
        """

        result = self._client._get(
            self.__class__.base_url(
                self.sys['id']
            )
        )

        self._update_from_resource(result)

        return self

    def delete(self):
        """
        Deletes the space
        """

        return self._client._delete(
            self.__class__.base_url(
                self.sys['id']
            )
        )

    def to_json(self):
        """
        Returns the JSON representation of the space.
        """

        result = super(Space, self).to_json()
        result.update({'name': self.name})
        return result

    def webhooks(self):
        """
        Provides access to webhook management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/webhooks

        :return: :class:`SpaceWebhooksProxy <contentful_management.space_webhooks_proxy.SpaceWebhooksProxy>` object.
        :rtype: contentful.space_webhooks_proxy.SpaceWebhooksProxy

        Usage:

            >>> space_webhooks_proxy = space.webhooks()
            <SpaceWebhooksProxy space_id="cfexampleapi">
        """
        return SpaceWebhooksProxy(self._client, self.id)

    def roles(self):
        """
        Provides access to role management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/roles

        :return: :class:`SpaceRolesProxy <contentful_management.space_roles_proxy.SpaceRolesProxy>` object.
        :rtype: contentful.space_roles_proxy.SpaceRolesProxy

        Usage:

            >>> space_roles_proxy = space.roles()
            <SpaceRolesProxy space_id="cfexampleapi">
        """
        return SpaceRolesProxy(self._client, self.id)

    def api_keys(self):
        """
        Provides access to api key management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys

        :return: :class:`SpaceApiKeysProxy <contentful_management.space_api_keys_proxy.SpaceApiKeysProxy>` object.
        :rtype: contentful.space_api_keys_proxy.SpaceApiKeysProxy

        Usage:

            >>> space_api_keys_proxy = space.api_keys()
            <SpaceApiKeysProxy space_id="cfexampleapi">
        """
        return SpaceApiKeysProxy(self._client, self.id)

    def preview_api_keys(self):
        """
        Provides access to preview api key management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/api-keys/preview-api-key/get-a-single-preview-api-key

        :return: :class:`SpacePreviewApiKeysProxy <contentful_management.space_preview_api_keys_proxy.SpacePreviewApiKeysProxy>` object.
        :rtype: contentful.space_preview_api_keys_proxy.SpacePreviewApiKeysProxy

        Usage:

            >>> space_preview_api_keys_proxy = space.preview_api_keys()
            <SpacePreviewApiKeysProxy space_id="cfexampleapi">
        """
        return SpacePreviewApiKeysProxy(self._client, self.id)

    def uploads(self):
        """
        Provides access to upload management methods.

        API reference: https://www.contentful.com/developers/docs/references/content-management-api/#/reference/uploads

        :return: :class:`SpaceApiKeysProxy <contentful_management.space_api_keys_proxy.SpaceApiKeysProxy>` object.
        :rtype: contentful.space_api_keys_proxy.SpaceApiKeysProxy

        Usage:

            >>> space_uploads_proxy = space.uploads()
            <SpaceUploadsProxy space_id="cfexampleapi">
        """
        return SpaceUploadsProxy(self._client, self.id)

    def environments(self):
        """
        Provides access to environment management methods.

        API reference: TBD

        :return: :class:`SpaceEnvironmentsProxy <contentful_management.space_environments_proxy` object.
        :rtype: contentful.environments_space_proxy.SpaceEnvironmentsProxy

        Usage:

            >>> space_environments_proxy = space.environments()
            <SpaceEnvironmentsProxy space_id="cfexampleapi">
        """

        return SpaceEnvironmentsProxy(self._client, self.id)

    def __repr__(self):
        return "<Space[{0}] id='{1}'>".format(
            self.name,
            self.id
        )
